#include <iostream>
#include <vector>
#include <set>
#include <cstring>

#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <fcntl.h>

#ifndef SOL_TCP
#  ifdef IPPROTO_TCP
#    define SOL_TCP IPPROTO_TCP
#  else
#    define SOL_TCP 6
#  endif /* IPPROTO_TCP */
#endif /* SOL_TCP */

#include <imsg.hh>
#include <message.hh>

#include <msgConstants.hh>

class iserver{
private:
	int masterSocket;
	std::set<int> socketSet;
	size_t maxConnections;
	bool timeLimit;
	unsigned long int inactivityTime, maxInactivityTime;
	
public:
	enum status{READ_EVENT, WRITE_EVENT, ERROR_EVENT};
	enum received{STAT_ERR, STAT_RECV};

	iserver():timeLimit(false), inactivityTime(0), maxInactivityTime(DEFAULT_SERVER_INACTIVITY_TIME*1000){}
	
	~iserver(){
		stopServer();
	}
	
	/**Create the server.
	 * Do the following:
	 * -* Create the address data structure
	 * -* Try to open the sock, set some option to this socket and finally bind it to a port
	 * -* Then start to listen to the incomming message
	 * -* Then treat all incoming message
	 */
	bool createServer(const std::string& addr, const std::string& port=DEFAULT_SERVER_PORT, size_t maxConnections=DEFAULT_MAX_CLIENT_CONNECTIONS){
		int result;
		this->maxConnections=maxConnections;

		struct addrinfo * address, *current;
		struct addrinfo hints;
		memset(&hints, 0, sizeof(struct addrinfo)); //Create empty hint
		hints.ai_family=AF_UNSPEC; //For both IPv4 and IPv6
		hints.ai_socktype=SOCK_STREAM;
		if(addr=="0.0.0.0" || addr.empty()){ //Bind the port on all possible address
			hints.ai_flags=AI_PASSIVE;
			result=getaddrinfo(NULL, port.c_str(), &hints, &address);
		}else{
			result=getaddrinfo(addr.c_str(), port.c_str(), &hints, &address);
		}
		if(result!=0){
			if(result==EAI_SYSTEM){
				int errnum=errno;
				std::cerr<<"Error while creating the address :"<<std::endl;
				std::cerr<<strerror(errnum)<<std::endl;
				return false;
			}else{
				std::cerr<<"Error while creating the address :"<<std::endl;
				std::cerr<<gai_strerror(result)<<std::endl;
				return false;
			}
		}
		
		for (current=address; current != NULL; current=current->ai_next) {
			masterSocket=socket(current->ai_family, current->ai_socktype, current->ai_protocol);
			if(masterSocket==-1){
				continue;
			}
			
			//Allow multiple connexions
			int reuse_addr=1;
			if(setsockopt(masterSocket, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, sizeof(reuse_addr))!=0){
				close(masterSocket);
				continue;
			}
			
			//Send message even if there are small
			int notcpdelay=1;
			if(setsockopt(masterSocket, SOL_TCP, TCP_NODELAY, &notcpdelay, sizeof(notcpdelay))!=0){
				close(masterSocket);
				continue;
			}

			if ((result=bind(masterSocket, current->ai_addr, current->ai_addrlen)) == 0)
				break;

			close(masterSocket);
		}

		if(current==NULL || masterSocket==-1 || result==-1){
			int errnum=errno;
			std::cerr<<"Error while creating socket :"<<std::endl;
			std::cerr<<strerror(errnum)<<std::endl;
			return false;
		}
		
		result=listen(masterSocket, maxConnections);
		if(result!=0){
			int errnum=errno;
			std::cerr<<"Cannot start to listen :"<<std::endl;
			std::cerr<<strerror(errnum)<<std::endl;
			return false;
		}
		
		freeaddrinfo(address);
		//Do not free 'current' because it points to a part of address that has just been freed
		
		return true;
	}
	
	std::vector< std::pair<status, int> > waitReadEvents() {
		std::vector< std::pair<status, int> > result;
		status status = ERROR_EVENT;
		
		listen(masterSocket, maxConnections);
		
		int highestFD=masterSocket;
		fd_set readfds;
		FD_ZERO(&readfds);
		FD_SET(masterSocket, &readfds);
		std::set<int>::const_iterator it;
		for(it=socketSet.begin(); it!=socketSet.end(); ++it){
			FD_SET(*it, &readfds);
			if(*it>highestFD){
				highestFD=*it;
			}
		}

		struct timeval timeout;
		timeout.tv_sec = 0;
		timeout.tv_usec = 1000; //Wake up every 1ms if nothing happened
		int newConnections = select(highestFD+1, &readfds, NULL, NULL, &timeout);

		if(newConnections < 0){
			std::cerr<<"Error on the socket while reading"<<std::endl;
			result.push_back(std::make_pair(status, -1));
			return result;
		}
		if(newConnections == 0){
			if(timeLimit){
				++inactivityTime;
				if(inactivityTime>=maxInactivityTime){
					result.push_back(std::make_pair(status, 0));
					return result;
				}
			}
		}else{
			if (FD_ISSET(masterSocket,&readfds)){
				if(!addClientConnection()){
					result.push_back(std::make_pair(status, -1));
					return result;
				}
			}
			
			inactivityTime=0;
			std::set<int>::const_iterator it;
			for(it=socketSet.begin(); it!=socketSet.end(); ++it){
				if(FD_ISSET(*it, &readfds)){
					status=READ_EVENT;
					result.push_back(std::make_pair(status, *it));
				}
			}
		}
		
		return result;
	}

	bool isMainSocket(int s){
		return(s==masterSocket);
	}

	void setInactivityTime(unsigned long int inactivityTime) {
		this->timeLimit=true;
		this->maxInactivityTime=inactivityTime*1000;
	}

	bool send(int sock, const message& data) {
		imsg msg;
		if(!msg.encode(data.data()))
			return false;
		ssize_t n;
		size_t tot = 0;
		size_t bleft = msg.length();
		while(tot < msg.length()) {
			n = ::send(sock, msg.data() + tot, bleft, MSG_NOSIGNAL);
			if(n < 1){
				std::cerr<<"Error while sending the message"<<std::endl;
				std::cerr<<"\t"<<strerror(errno)<<std::endl;
				break;
			}
			tot += n;
			bleft -= n;
		}
		return (tot == msg.length());
	}

	std::pair<received, message> receive(int sock){		
		message result;
		imsg msg;
		ssize_t res;
		char header[msg.hsize()];
		res = ::recv(sock, header, msg.hsize(), 0);
		if(res<=0){
			removeClientConnection(sock);
			return std::make_pair(STAT_ERR, result);
		}
		if(res != static_cast<ssize_t>(msg.hsize())){
			removeClientConnection(sock);
			return std::make_pair(STAT_ERR, result);
		}
		msg.decodeHeader(header);
		char body[msg.blength()];
		res = ::recv(sock, body, msg.blength(), MSG_WAITALL);
		if(res != static_cast<ssize_t>(msg.blength())){
			removeClientConnection(sock);
			return std::make_pair(STAT_ERR, result);
		}
		msg.decodeBody(body);
		result.decode(msg.body());
		return std::make_pair(STAT_RECV, result);
	}

	void stopServer() {
		close(masterSocket);
	}

private:
	bool addClientConnection(){
		int connection=accept(masterSocket, NULL, NULL);
				
		if(connection<0){
			std::cerr<<"Could not accept new connection"<<std::endl;
			return false;
		}
		
		if(socketSet.size()<maxConnections){
			socketSet.insert(connection);
			message ackMsg;
			ackMsg.encode(DEFAULT_SERVER_NAME, SERVER_CONNECTION_ACK);
			send(connection, ackMsg);
		}else{
			std::cerr<<"Maximum number of connections reached, reject the new one"<<std::endl;
			message errorMsg;
			errorMsg.encode(DEFAULT_SERVER_NAME, SERVER_CONNECTION_NACK);
			send(connection, errorMsg);
			return false;
		}
		
		return true;
	}
	
	void removeClientConnection(int toClose){
		if(socketSet.find(toClose)!=socketSet.end()){
			::close(toClose);
			socketSet.erase(toClose);
		}
	}
};
